import XME,time,numpy as np,traceback
try:
	import _pickle as pickle
except:
	try:
		import cPickle as pickle
	except:
		import pickle
from multiprocessing import cpu_count,Pipe,Manager
from threading import Thread

ALL_PROCESSES="<XME-MPI-MARK: ALL PROCESSES>"
ANY_PROCESS="<XME-MPI-MARK: ANY PROCESS>"
TASK_END="<XME-MPI-MARK: TASK END>"
STATUS_BEGIN="<XME-MPI-STATUS: BEGIN>"
STATUS_IDLE="<XME-MPI-STATUS: IDLE>"
STATUS_ALLOCATED="<XME-MPI-STATUS: ALLOCATED>"
STATUS_INRUN="<XME-MPI-STATUS: INRUN>"
STATUS_END="<XME-MPI-STATUS: END>"
TASK_NO_REGISTERED="<XME-MPI-RUN: TASK_NO_REGISTERED>"
TASK_RUN_FAILURE="<XME-MPI-RUN: TASK_NO_REGISTERED>"
BUFFER_EMPTY="<XME-MPI-BUFFER: EMPTY>"
BUFFER_EOF="<XME-MPI-BUFFER: EOF"
def get_process_mark(tpid):
	if tpid==0: return "<XME-MPI-MARK: MAIN PROCESS>"
	else: return f"<XME-MMPI-MARK: TASK PROCESS {tpid}>"
class MPI:
	def __init__(self,mainfun,**xmeargs):
		self.Tasks=Manager().dict()
		self.buffersize=XME.get_par(xmeargs,"mpi_buffer_size",0xff)
		self.buffer=Manager().list([BUFFER_EMPTY for i in range(self.buffersize)])
		self.bufferpos=0

		self.mainfun=mainfun
		self.xmeargs=xmeargs
		self.pnum=XME.get_par(xmeargs,"pnum",cpu_count())
		if self.pnum>cpu_count():
			self.xmeargs["pnum"]=cpu_count()
			self.pnum=cpu_count()
		elif self.pnum<2: 
			raise ValueError("XME-MPI need at least 2 processes (1 main process + 1 task process)!")
			exit(0)

	def run(self,*args,**kwargs):
		m=Manager()
		status=m.dict()
		# pnum : task_processes: pnum-1
		Main_Connector={} #pipes to main process
		Task_Connector=[] #pipes to tasks processes
		tpids=[]
		for i in range(1,self.pnum):
			tpid=get_process_mark(i)
			tpids.append(tpid)
			status[tpid]=STATUS_BEGIN
			pc,cc=Pipe()
			Main_Connector[tpid]=pc
			Task_Connector.append(cc)
		xme=XME.XME(self.__task__,self.mainfun,**self.xmeargs)
		kwargs.update({"XMEMPI":(self,status,Main_Connector)})
		return xme.gfun(status,xme.Array(tpids),xme.Array(Task_Connector),garg=args,gargs=kwargs)[0][0]

	def __str__(self): return f"<class XME::MPI @ {hex(id(self))}: Task processes: {self.pnum}; Registed Tasks: {len(self.Tasks)}>"
	def __setitem__(self,key,value): self.Tasks[key]=value
	def __getitem__(self,key): return self.Tasks[key]

	def acquire(self, status, connets, funid, args=(), kwargs={}, to=ANY_PROCESS,block=False):
		def send_and_recv(to,pos):
			connets[to].send(self.__encode__((funid,args,kwargs)))
			self.buffer[pos]=connets[to].recv()
		if to==ALL_PROCESSES: return [self.acquire(status, connets, funid, args, kwargs,tpid) for tpid in status.keys()]
		elif type(to) in (list,tuple): return [self.acquire(status, connets, funid, args, kwargs,tpid) for tpid in to]
		elif to==ANY_PROCESS:
			closed=[]
			while True:
				for tpid,sta in status.items():
					if sta==STATUS_IDLE:
						status[tpid]=STATUS_ALLOCATED
						pos=self.bufferpos
						if not block: Thread(target=send_and_recv,args=(tpid,pos)).start()
						else: send_and_recv(tpid,pos)
						self.bufferpos+=1
						if self.bufferpos==self.buffersize: self.bufferpos=0
						return pos
					elif sta==STATUS_END and tpid not in closed: closed.append(tpid)
					if set(status.keys())==set(closed):return
		else:
			while True:
				if status[to]==STATUS_IDLE:
					status[to]=STATUS_ALLOCATED
					pos=self.bufferpos
					if not block: Thread(target=send_and_recv,args=(to,pos)).start()
					else: send_and_recv(tpid,pos)
					self.bufferpos+=1
					if self.bufferpos==self.buffersize: self.bufferpos=0
					return pos
				elif status[to]==STATUS_END:return

	def get(self,pos, timeout=0,looptime=0.1):
		if pos==None or pos>=self.buffersize: return BUFFER_EOF
		t0=time.time()
		while self.buffer[pos]==BUFFER_EMPTY:
			if timeout!=0 and time.time()-t0>timeout: 
				print(f"Warning: timeout for getting result from buffer at position {pos}")
				return
			time.sleep(looptime)
		result=self.__decode__(self.buffer[pos])
		self.buffer[pos]=BUFFER_EMPTY
		return result

	def close(self, status, connets, to=ALL_PROCESSES):
		if to==ALL_PROCESSES:
			closed=[]
			while True:
				for tpid,sta in status.items():
					if tpid not in closed:
						if sta==STATUS_IDLE:
							connets[tpid].send(self.__encode__(TASK_END))
							closed.append(tpid)
						elif sta == STATUS_END: closed.append(tpid)
					if set(status.keys())==set(closed):return
		else:
			if status[to] == STATUS_END: return
			while True:
				if status[to]==STATUS_IDLE:
					connets[to].send(self.__encode__(TASK_END))
					return

	def __encode__(self,data): return pickle.dumps(data)

	def __decode__(self,data): return pickle.loads(data)

	def __task__(self, status, tpid, connect, print=print):
		while True:
			try:
				status[tpid]=STATUS_IDLE
				request=self.__decode__(connect.recv())
				status[tpid]=STATUS_INRUN
				#request message: (funname, arglist, kwargdict)
				if request==TASK_END: 
					status[tpid]=STATUS_END
					break
				funid, arg, kwargs=request
				if funid in self.Tasks: 
					result=self.Tasks[funid](*arg,**kwargs)
					connect.send(self.__encode__(result))
				else: connect.send(self.__encode__(TASK_NO_REGISTERED))
			except Exception as err:
				traceback.print_exc()
				connect.send(self.__encode__(TASK_RUN_FAILURE))


from multiprocessing import cpu_count
from .XMElib.ArrayOperator import ArrayOperator
from .XMElib.Executor import Executor
from .XMElib.Logputter import Logputter

def get_par(args,name,default=None):
	try: return args[name]
	except: return default

class XME:
	aoobj_array=[]
	exobj_array=[]
	class Array:
		def __init__(self,array):
			self.array=array
			self.length=len(array)
		def __len__(self): return self.length
		def __str__(self): return str(self.array)
		def __getitem__(self,key): return self.array[key]
		def __delitem__(self,key): del self.array[key]
		def __setitem__(self,key,value): self.array[key]=value

	def __init__(self,*fun,**args):
		if get_par(args,"pnum",cpu_count())>cpu_count(): args["pnum"]=cpu_count()
		elif get_par(args,"pnum",cpu_count())<1: args["pnum"]=1
		self.pnum=get_par(args,"pnum",cpu_count())
		self.funs=[]
		if get_par(args,"do_with_log",True):
			self.logobj=Logputter(get_par(args,"logfile"),get_par(args,"show_version_info",False))
			self.logobj.print_in_screen=get_par(args,"print_in_screen",True)
		else: self.logobj=None
		def func(funum,*targ,**args):
			args.update({"logobj":self.logobj})
			if self.logobj!=None: args.update({"print":self.logobj.write_log})
			calnum=get_par(args,"calnum",0)
			if calnum==0:
				for i in targ:
					if type(i)==self.Array: calnum=max(calnum,i.length)
				for i in args.keys():
					if type(args[i])==self.Array: calnum=max(calnum,args[i].length)
			if calnum==0: calnum+=1 #at least run once
			ao=self.ao(calnum,self.pnum)
			for i in targ: #first set
				if type(i)!=self.Array: ao.add_common_args(i)
				else: ao.add_argscut(i.array)
			for i in fun[funum].__code__.co_varnames: #follow sequence
				if i in args.keys():
					if type(args[i])!=self.Array: ao.add_common_args(args[i])
					else: ao.add_agrscut(args[i].array)
			return ao
		if len(fun)>0:
			def single_fun(*targ,**args):
				ao=func(0,*targ,**args)
				ex=self.ex(fun[0],pnum=self.pnum)
				ex.build_from_ao(ao)
				ao.result_combine()
				return ao.results
			self.fun=single_fun
			def multi_funs(funum_array=range(len(fun)),targ_array=[[]]*len(fun),args_array=[{}]*len(fun)):
				if len(funum_array)!=len(targ_array) or len(funum_array)!=len(args_array):
					print("Error parameters number")
					return ()
				ao=[]
				results=[]
				tfuns=[]
				for i in range(len(funum_array)):
					tfuns.append(fun[funum_array[i]])
					ao.append(func(funum_array[i],*(targ_array[i]),**(args_array[i])))
				ex=self.ex(*tfuns,pnum=self.pnum)
				ex.build_from_ao(ao)
				results=[]
				for i in range(len(funum_array)):
					ao[i].result_combine()
					results.append(ao[i].results)
				return tuple(results)
			self.funs=multi_funs
			def gfun(*farg,garg=[],gargs={},**fargs): return multi_funs([-1,0],targ_array=[garg,farg],args_array=[gargs,fargs])
			self.gfun=gfun
			def gfuns(funum_array=range(len(fun)-1),farg_array=[[]]*(len(fun)-1),fargs_array=[{}]*(len(fun)-1),garg=[],gargs={}): return multi_funs([-1]+funum_array,targ_array=[garg]+farg_array,args_array=[gargs]+fargs_array)
			self.gfuns=gfuns
	def ao(self,calnum,pnum=None):
		if pnum==None: pnum=self.pnum
		self.aoobj_array.append(ArrayOperator(cal_num=calnum,pnum=pnum))
		return self.aoobj_array[-1]
	def ex(self,*fun,**args):
		self.exobj_array.append(Executor(*fun,**args))
		return self.exobj_array[-1]
	def clean(self):
		self.aoobj_array=[]
		self.exobj_array=[]
def build(*fun,**args):
	return XME(*fun,**args)



from multiprocessing.managers import BaseManager
class __Manager__(BaseManager): pass
def __get_manager__():
	m=__Manager__()
	m.start()
	return m

def Object(class_name,class_type,*initargs,**initkwargs):
	__Manager__.register(class_name,class_type)
	return eval(f"__get_manager__().{class_name}(*initargs,**initkwargs)")


from multiprocessing import cpu_count,Manager
def get_par(args,name,default=None):
	try:
		return args[name]
	except:
		return default
class ArrayOperator:
	def __init__(self,**args):
		self.pnum=get_par(args,"pnum",cpu_count())
		self.results=[]
		self.args=[]
		self.cal_num=get_par(args,"cal_num",0)
		manager=Manager()
		for i in range(self.pnum):
			self.results.append(manager.list())
			self.args.append([])
		pn=0
		for i in range(self.cal_num):
			self.args[pn].append([])
			pn+=1
			if pn==self.pnum: pn=0
	def add_argscut(self,args,default_args=None):
		pnum=0
		co=0
		cal_n=0
		for i in args:
			if cal_n<self.cal_num: self.args[pnum][co].append(i)
			pnum+=1
			cal_n+=1
			if pnum==self.pnum:
				pnum=0
				co+=1
		for i in range(cal_n,self.cal_num):
			self.args[pnum][co].append(default_args)
			pnum+=1
			if pnum==self.pnum:
				pnum=0
				co+=1
	def add_common_args(self,*args):
		if len(args)>1:
			for i in args: self.add_common_args(i)
		else:
			pnum=0
			co=0
			for i in range(self.cal_num):
				self.args[pnum][co].append(args[0])
				pnum+=1
				if pnum==self.pnum:
					pnum=0
					co+=1
	def result_combine(self):
		new_result=[]
		for j in range(len(self.results[0])):
			for i in range(len(self.results)):
				try: new_result.append(self.results[i][j])
				except: continue
		self.results=tuple(new_result)


#Monitor.py
__version__="1.0"
__author__="WacmkXiaoyi@gmail.com"
__doc__="""
Monitor: Multi process/thread message monitor

Basic data atom: **Message

package needs: json
"""
import copy,json
from multiprocessing import Manager

class Monitor:
	security=None
	def __init__(self,security=None,onlock=lambda op:True,unlock=lambda op:True):
		"""
		update message
		args: 	security 
				onlock (fun), default disable (True)
				unlock (fun), default disable (True)
		"""
		self.__Message__=Manager().dict()
		if security!=None:
			self.security=security
			onlock=security.acquire
			unlock=security.release
		self.update_onlock(onlock)
		self.update_unlock(unlock)
	def __str__(self):
		return self.get_str()
	def get_str(self):
		str00=""
		return f"<class Monitor @ {hex(id(self))}: <Message: {self.__Message__}>; {str00 if self.security==None else self.security.get_str()}>"
	def __getitem__(self,key):
		return self.get(key)
	def get(self,*keys):
		'''
		if self.__onlock__("Reader"): 
			if len(keys)==0:result=self.__Message__
			else: 
				result={}
				for key,value in self.__Message__.items(): result[key]=value
		self.__unlock__("Reader")
		return result
		'''
		if len(keys)==0:result=copy.deepcopy(self.__Message__)
		elif len(keys) ==1 : result=self.__valuetable__[keys[0]]
		else: 
			result={}
			for key,value in self.__Message__.items():
				if key in keys: result[key]=value
		return result
	def items(self):
		'''
		if self.__onlock__("Reader"): result=self.__Message__.items()
		self.__unlock__("Reader")
		return result
		'''
		return self.__Message__.items()
	# lock update
	def update_onlock(self,onlock):
		self.__onlock__=onlock
	def update_unlock(self,unlock):
		self.__unlock__=unlock

	def update(self,newmessage, Operator="Undef"):
		"""
		update message
		args: 	message (dict)
				operator
		"""
		#if Operator=="Reader": return
		if self.__onlock__(Operator): self.__Message__.update(newmessage)
		self.__unlock__(Operator)
	def delete(self,*keys,Operator="Undef"):
		"""
		Delete message by key, default by all keys
		args: 	message (dict)
		"""
		#if Operator=="Reader": return
		if len(keys)==0: keys=self.__Message__.keys()
		if self.__onlock__(Operator): 
			for key in keys: del self.__Message__[key]
		self.__unlock__(Operator)
		
	def __compact_data__(self,b_msg):
		def convert(char):
			char=bin(char)[2:]
			char="0"*((8-len(char)%8)%8)+char
			char=int(char[int(len(char)/2):]+char[:int(len(char)/2)],2)
			return char
		return bytes([convert(cr) for cr in b_msg])
	def marshal(self):
		"""
		Serialize and Marshal to bytes:
		dict -> serialize => string -> encode() => UTF-8 list -> compact() => bytes group
		"""
		'''
		if self.__onlock__("Reader"):result=self.__compact_data__(json.dumps(self.__Message__).encode("utf-8"))
		self.__unlock__("Reader")
		return result
		'''
		return self.__compact_data__(json.dumps(self.__Message__).encode("utf-8"))
	def unmarshal(self,m_message):
		"""
		Unmarshal from Bytes and deserialize
		args: 	marshal message (string)
		"""
		m_message=self.__compact_data__(m_message).decode("utf-8")
		self.update(json.loads(m_message))
if __name__=="__main__":
	def debug():
		mm=Monitor()
		mm.update({"a":str(lambda:"a"),"b":str(lambda x:"b"*x)})
		print(mm,"\n")

		ma=mm.marshal()
		print(ma,"\n")

		mb=Monitor()
		mb.unmarshal(ma)
		print(mb["a"],mb["b"])
	debug()


#Monitor.py
__version__="1.0"
__author__="WacmkXiaoyi@gmail.com"
__doc__="""
Class Security: shared data operations protocol (SecProto)

Package Needs: time

All types of locks only allow one thread to operate on shared data
"""

import time,random
from multiprocessing import Manager

#LockType:
DISABLE=0 #disable lock, Please use built-in lock in module threading!!!!
MUTEX=1 #mutex lock
AUTHORISATION=2 #authorisation operators
READWRITE=3 #Readwrite lock
SPIN=4 #Spin Lock

DEADLOCKMARK_ONLOCK="Onlock failure: illegal operator or deadlock!"
DEADLOCKMARK_UNLOCK="Unlock failure: illegal operator or deadlock!"

class ADMIN:
	def __str__(self):
		return "<SecProto::Operator:ADMIN>"

class Security:
	def __default_Deadlockfun_onlock__(self): return DEADLOCKMARK_ONLOCK
	def __default_Deadlockfun_unlock__(self): return DEADLOCKMARK_UNLOCK
	def __init__(self,LockType=READWRITE):
		__manager__=Manager()
		self.set_lock(__manager__.RLock(),__manager__.RLock())
		#Lock status, Locker, Locktype, Spintime, Default Wait Time
		self.ADMIN=ADMIN()
		self.__status__=__manager__.list([False,self.ADMIN,LockType,0.1,0,self.__default_Deadlockfun_onlock__,self.__default_Deadlockfun_unlock__])
		self.__AuthMap__=__manager__.list([self.ADMIN])
		self.update_lock(LockType)

	def get_str(self):
		def getlocktype():
			return {
				DISABLE:"disable",
				MUTEX:"Mutex",
				AUTHORISATION:"Authorisation",
				SPIN:"Spin",
				READWRITE:"Read-write"
			}[self.__status__[2]]
		def getstatus():
			return{
				True:f"Onlock (Type: {getlocktype()}), Locker: {self.__status__[1]}",
				False:"Unlock"
			}[self.__status__[0]]
		return f"<SecProto (class Security) @ {hex(id(self))}: <status: {getstatus()}>>"
	def __str__(self):
		return self.get_str()
	def set_lock(self,lockpoc,lockaut):
		self.__poclock__=lockpoc
		self.__auxlock__=lockaut #aux lock
		
	def update_lock(self,LockType):
		"""
		Update Locktype, and reallocate the lockfun
		args: LockType
		"""
		#self.__poclock__.acquire()
		self.__status__[2]=LockType
		self.__LockFun__={
			DISABLE:lambda Operator:True,
			MUTEX:self.__Mutex__,
			AUTHORISATION:self.__Auth__,
			SPIN:self.__Spin__,
			READWRITE:self.__ReadWrite__
		}[LockType]
		self.__UnLockFun__={
			DISABLE:lambda Operator:True,
			MUTEX:self.__Mutex_Release__,
			AUTHORISATION:self.__Auth_Release__,
			SPIN:self.__Spin_Release__,
			READWRITE:self.__ReadWrite_Release__
		}[LockType]
		#self.__poclock__.release()

	#Spin Lock 
	def __Spin__(self, Operator):
		return self.__Mutex__(Operator)
	def __Spin_Release__(self, Operator):
		return self.__Mutex_Release__(Operator)

	#MUTEX LOCK (Recommond)
	
	def __Mutex__(self,Operator):
		if self.__status__[0]: return self.__status__[1] == Operator
		# Try to acquire the lock
		acquired = self.__poclock__.acquire()
		if acquired:
			# Lock acquired successfully
			self.__status__[0] = True
			self.__status__[1] = Operator
			return True
		return False

	def __Mutex_Release__(self, Operator):
		# Release the lock only if the current operator matches
		if self.__status__[0] and self.__status__[1] == Operator:
			# Release the lock
			self.__status__ [0] = False
			self.__status__ [1] = None
			try:
				self.__poclock__.release()
			except:
				pass
			return True
		return False

	def MuteTime(self,mutetime):
		self.__status__[3]=mutetime


	#AUTHORISATION Lock
	#Only Operator in __AuthMap__ can Change Data
	def authorize(self,*Operator):
		"""
		Authorize who can access data
		args: Operator
		"""
		[self.__AuthMap__.append(i) for i in Operator]
	def __Auth__(self,Operator):
		if Operator not in self.__AuthMap__: return False
		return self.__Mutex__(Operator)
	def __Auth_Release__(self, Operator):
		if Operator not in self.__AuthMap__: return False
		return self.__Mutex_Release__(Operator)

	
	#Readwrite lock
	def registe(self,Operator):
		"""
		Registe who (only one) can access data
		args: Operator
		"""
		self.__status__[1]=Operator
	def __ReadWrite__(self, Operator):
		if Operator!=self.__status__[1]: return False
		return self.__Mutex__(Operator)
	def __ReadWrite_Release__(self, Operator):
		if Operator!=self.__status__[1]: return False
		return self.__Mutex_Release__(Operator)

	def WaitTime(self,waittime):
		self.__status__[4]=waittime
	def DeadLockFun(self,acquire=None,release=None):
		if acquire!=None: self.__status__[5]=acquire
		if release!=None: self.__status__[6]=release
	def acquire(self,Operator,WaitTime=None,DeadLockFun=None):
		"""
		Determine whether to lock, and if it has already been locked by the current operator, do not take any action
		args: 	1. Operator
				2. WaitTime: if timer>WaitTime, call DeadLockFun() [option, default 0 -> no check]
				3. DeadLockFun [option, default do nothing]

		Please set `WaitTime` to avoid death lock

		Return bool:
		Ture : Onlock Successful 
		False: Onlock failure (AUTHORISATION & READWRITE only)
		
		Recommond Usage:
		1. sp.acquire(op[,WaitTime]) MUTEX & SPIN
		2. if sp.acquire(op[,WaitTime]): <...> AUTHORISATION & READWRITE
		"""
		#Check Disable
		'''
		if self.__status__[2] == DISABLE: return True 

		#check ReadWrite
		elif self.__status__[2] == READWRITE: return self.__LockFun__(Operator)

		#Check illegal operator in AUTHORISATION
		elif self.__status__[2] == AUTHORISATION and Operator not in self.__AuthMap__:return False

		#Check Onlock
		elif Operator==self.__status__[1]: return True
		'''
		#Legitimate operator
		if WaitTime==None: WaitTime=self.__status__[4]
		if DeadLockFun==None: DeadLockFun=self.__status__[5]
		time1=time.time()
		while not self.__LockFun__(Operator):
			if WaitTime!=0 and time.time()-time1>WaitTime: return DeadLockFun() #Detecting death lock
			if self.__status__[2] == MUTEX: time.sleep(self.__status__[3])

		return True # onlock successfully

	def release(self,Operator,WaitTime=None,DeadLockFun=None):
		"""
		The release operation of a lock, any lock (except DISABLE and READWRITE) that is not in use after being used should be released
		args: Operator

		Return bool
		True: Unlock successfully (including unlock status...)
		False: Unlock failure, e.g., illegal operator
		"""
		'''
		if self.__status__[2]==DISABLE: return True
		elif self.__status__[1]!=Operator: return False
		elif self.__status__[2]==READWRITE: return True
		self.__LockFun__(Operator)
		return not self.__status__[0]
		'''
		if WaitTime==None: WaitTime=self.__status__[4]
		if DeadLockFun==None: DeadLockFun=self.__status__[6]
		time1=time.time()
		while not self.__UnLockFun__(Operator):
			if WaitTime!=0 and time.time()-time1>WaitTime: return DeadLockFun()
		return True

if __name__=="__main__":
	a=0
	def debug(LockType):
		import threading,hashlib
		#initial debug...
		def testsp(Operator):
			global a
			time1=time.time()
			print("Operator:",Operator,"\n\tStatus->Unlock:",sp,"\n\tValue:",a,"\n")
			if sp.acquire(Operator):
				print("Operator:",Operator,"\n\tStatus->Onlock:",sp,"\n\tValue:",a,"\n")
				for i in range(10000000): a+=1
			else: print("Operator:",Operator,"\n\tOnlock failure:",sp)
			if sp.release(Operator): print("Operator:",Operator,"\n\tStatus->Release lock:",sp,"\n\tValue:",a,"\n")
			else: print("Operator:",Operator,"\n\tUnlock failure:",sp)
			time2=time.time()
			print("Operator",Operator,": finished, use time:",time2-time1,"\n")
		def get_thread_name(name,size=8):
			return "Thread-"+hashlib.md5(name.encode('utf-8')).hexdigest()[:size]
		global a
		a=0 #ideal result=40000000 (disable) 20000000(other)
		sp=Security(LockType)
		name1=get_thread_name("th1")
		th1=threading.Thread(target=testsp,args=(name1,),name=name1)
		name2=get_thread_name("th2")
		th2=threading.Thread(target=testsp,args=(name2,),name=name2)
		sp.authorize(name1)
		sp.registe(name2)
		th1.start()
		th2.start()
		th1.join()
		th2.join()
	debug(4)


from multiprocessing import Pool,cpu_count#,Lock
import numpy
def tuplize(array):
	if type(array) in (list,tuple, numpy.ndarray):
		result=[]
		for i in array:
			if type(i) in (list, numpy.ndarray): result.append(tuplize(i))
			else: result.append(i)
		return tuple(result)
	return array
def get_par(args,name,default=None):
	try: return args[name]
	except: return default
class Executor:
	default_fun_strc="fun"
	def __init__(self,*fun,**args):
		self.clear_fun()
		self.add_fun(fun)
		self.pnum=get_par(args,"pnum",cpu_count())
		self.results=[]
	def do_fun(self,args=(),funname=None,returns=None):
		for i in args:
			if funname==None: funname=tuple(self.fun.keys())[-1]
			result=self.fun[funname](*i)
			if returns!=None: returns.append(result)
		if returns!=None: return returns
	def add_a_pool(self,args,funname=None,close=True,join=True,returns=None):
		pool=Pool(1)
		if funname==None: funname=tuple(self.fun.keys())[-1]
		result=pool.apply_async(self.do_fun,(args,funname,returns))
		if close: pool.close()
		if join: pool.join()
		return tuplize(result.get())
	def build_from_ao(self,ao,close=True,join=True,pnum=None):
		if pnum==None: pnum=self.pnum
		pool=Pool(pnum)
		results=[]
		infunname=tuple(self.fun.keys())
		if type(ao) is tuple or type(ao) is list:
			funname=[]
			for i in range(len(ao)):
				if i < len(infunname): funname.append(infunname[i])
			for i in range(len(infunname),len(ao)): funname.append(infunname[-1])
			new_ao={}
			for i in funname: new_ao[i]=()
			for i in range(len(funname)): new_ao[funname[i]]+=(ao[i],)
			ao=new_ao
		elif type(ao) is dict:
			new_ao={}
			for i in ao.keys():
				if i in infunname:
					if type(ao[i]) is tuple or type(ao[i]) is list: new_ao[i]=tuple(ao[i])
					else: new_ao[i]=(ao[i],)
				else: print("Warnning: fun name:",i,"no in library")
			ao=new_ao
		else: ao={infunname[-1]:(ao,)}
		for i in ao.keys():
			for j in ao[i]:
				for k in range(j.pnum): results.append(pool.apply_async(self.do_fun,(j.args[k],i,j.results[k])))
		if close: pool.close()
		if join: pool.join()
		new_results=[]
		for i in results:
			new_results.append(i.get())
		return new_results
	def add_fun(self,fun):
		if type(fun) is tuple or type(fun) is list:
			tvi=0
			for i in range(len(self.fun.keys()),len(self.fun.keys())+len(fun)):
				self.fun[self.default_fun_strc+str(i)]=fun[tvi]
				tvi+=1
		elif type(fun) is dict:
			for i in self.fun.keys(): self.fun[i]=fun[i]
		else: self.fun[self.default_fun_strc+str(len(self.fun.keys()))]=fun
	def clear_fun(self): self.fun={}


import setuptools
 
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
 
setuptools.setup(
    name="XME",
    version="4.1.3",
    author="Junxiang Huang & Weihui Li",
    author_email="huangjunxiang@mail.ynu.edu.cn",
    description="This is a Python interface for process pool establishment, and result collection based on the Multiprocessing module.",
    long_description=long_description, 
    long_description_content_type="text/markdown",
    url="https://github.com/wacmkxiaoyi/Xenon-Multiprocessing-Engine", 
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',  
)


import datetime
import os
from XME.XMElib import XME_Version
def get_par(args,name,default=None):
	try:
		return args[name]
	except:
		return default
class Logputter:
	logfile=None
	time=True
	msgtype=True
	default_msgtype="Message"
	processing=True
	indent=True
	indent_num=4
	indent_type="-"
	print_in_screen=True
	def __init__(self,logfile=None,show_version_info=True):
		self.logfile=logfile
		self.write_header(show_version_info)
	def write_header(self,show_version_info=True):
		strc="==============================================\n"
		strc+="========Xenon  Multiprocessing  Engine========\n"
		strc+="=================WACMK   Tech=================\n"
		strc+=f"================Version {XME_Version}=================\n"
		strc+="==============================================\n"
		if self.print_in_screen and show_version_info: print(strc)
		if self.logfile!=None:
			file=open(self.logfile,"a")
			file.writelines(strc+"\n")
			file.close()
	def write_log(self,*message,**args):
		strc=""
		msgtype=get_par(args,"msgtype")
		indent=get_par(args,"indent",0)
		pid=get_par(args,"pid",os.getpid())
		if msgtype==None: msgtype=self.default_msgtype
		if self.msgtype and not msgtype: strc+="("+msgtype+")"
		if self.time: strc+=datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
		if self.processing and pid!=None: strc+="(pid@"+str(pid)+")"
		if strc!="": strc+=":"
		if self.indent:
			for i in range(indent):
				for j in range(self.indent_num): strc+=self.indent_type
		for i in message: strc+=str(i)+" "
		if self.print_in_screen: print(strc)
		if self.logfile!=None:
			file=open(self.logfile,"a")
			file.writelines(strc+"\n")
			file.close()
if __name__=="__main__":
	logobj=Logputter("test.log")
	logobj.write_log("test_log")



import multiprocessing
from .Object import Object
from .Monitor import Monitor
from .Security import Security
from copy import deepcopy
import hashlib,os,time,traceback,threading


GUARD_FUN_FAIL= "<-Guard Function Executes Failure->"
XMEM_GET_FAIL= "<-XMEM GET Failure->"
class XMEManager:
	def __init__(self,sec,mon,initialtable={},size=32):
		__manager__=multiprocessing.Manager()
		initialtable["__XMEM__"]={
			"Lock":{
				"initial":self.initial_lock,
				"update_onlock":self.update_onlock,
				"update_unlock":self.update_unlock,
				"update":self.update_lock
			},
			"Table":{
				"update":self.update,
				"set":self.set,
				"delete":self.delete,
				"items":self.items,
				"get":self.get_table
			},
			"Monitor":{
				"get":self.get_msg,
				"update":self.update_msg,
				"delete":self.delete_msg,
				"copy_msg":self.copy_msg
			}
		}
		initialtable["__GUARD__"]=__manager__.dict({"Status": True})
		initialtable["__STATUS__"]=__manager__.dict()
		fpc={}
		for key1,value1 in initialtable["__XMEM__"].items():
			for key2,value2 in value1.items(): fpc[f"{key1}::{key2}"]=value2
		initialtable["__XMEM__"].update(fpc)
		self.__valuetable__=__manager__.dict(initialtable)
		self.__Buffer_Call__=__manager__.dict({"GUARD":__manager__.list()})
		self.op_size=size
		self.security=sec
		self.monitor=mon
		self.update_onlock(self.security.acquire)
		self.update_unlock(self.security.release)

	def get_str(self):
		return f"<Class XMEManager @ {hex(id(self))} (Operator: {self.AllocOperator()}): <Values: {str(self.__valuetable__)}>; {self.security.get_str()}; {self.monitor.get_str()}>"
	def __str__(self):
		return self.get_str()

	# Buffer operation, usually call by Guard function
	def append(self,fun,Operator=None,**kwargs):
		#args,fun,
		if Operator==None: Operator=self.AllocOperator()
		self.__Buffer_Call__["GUARD"].append((fun,Operator,kwargs))
	def exec(self):
		Operator=self.__Buffer_Call__["GUARD"][0][1]
		if self.__Buffer_Call__["GUARD"][0][0]==__XMEHJ__:
			self.__valuetable__["__STATUS__"][Operator]=self.__Buffer_Call__["GUARD"][0][2]
			del self.__Buffer_Call__["GUARD"][0]
			return
		try:
			for key,value in self.__valuetable__["__XMEM__"]["Table"].items():
				if self.__Buffer_Call__["GUARD"][0][0]==value and "Operator" in value.__code__.co_varnames and Operator not in self.__Buffer_Call__["GUARD"][0][2]: 
					self.__Buffer_Call__["GUARD"][0][2]["Operator"]=self.security.ADMIN
					break
			for key,value in self.__valuetable__["__XMEM__"]["Monitor"].items():
				if self.__Buffer_Call__["GUARD"][0][0]==value and "Operator" in value.__code__.co_varnames and Operator not in self.__Buffer_Call__["GUARD"][0][2]: 
					self.__Buffer_Call__["GUARD"][0][2]["Operator"]=self.monitor.security.ADMIN
					break
			result=self.__Buffer_Call__["GUARD"][0][0](**self.__Buffer_Call__["GUARD"][0][2])
		except Exception as e:
			traceback.print_exc()
			result=GUARD_FUN_FAIL
		del self.__Buffer_Call__["GUARD"][0]
		self.__Buffer_Call__[Operator]=result
	def clean(self,Operator=None):
		if Operator==None: Operator=self.AllocOperator()
		try: del self.__Buffer_Call__[Operator]
		except: pass
	def get(self,Operator=None,MaxWaitTime=0,LoopTime=0.1):
		if Operator==None: Operator=self.AllocOperator()
		time1=time.time()
		while (Operator not in self.__Buffer_Call__): 
			if MaxWaitTime!=0 and time.time()-time1>MaxWaitTime: return XMEM_GET_FAIL
			time.sleep(LoopTime)
		result=self.__Buffer_Call__[Operator]
		del self.__Buffer_Call__[Operator]
		return result
	def status(self,Operator=None,status=None):
		if status==None: status=f"<-HeartJump::{time.time()}->"
		self.append(__XMEHJ__,Operator,status=status)
	def get_buffers(self):
		return deepcopy(self.__Buffer_Call__)
	def len(self):
		return len(self.__Buffer_Call__["GUARD"])
	def __len__(self):
		return self.len()

	# Lock operation functions, no suggest to call in manual!
	def update_onlock(self,onlock):
		self.__onlock__=onlock
	def update_unlock(self,unlock):
		self.__unlock__=unlock
	def update_lock(self,LockType):
		self.security.update_lock(LockType)
		self.monitor.security.update_lock(LockType)
	def initial_lock(self,Locks):
		self.security.set_lock(Locks[0],Locks[1])
		self.monitor.security.set_lock(Locks[2],Locks[3])
	
	#Table Operation
	def get_table(self,*keys):
		'''
		if self.security.acquire("Reader"):
			if len(keys)==0:  result=self.__valuetable__
			else:  #no modifiable!
				result={}
				for key,value in self.__valuetable__.items(): result[key]=value
		self.security.release("Reader")
		return result
		'''
		if len(keys)==0:  result=deepcopy(self.__valuetable__)
		elif len(keys) ==1 : result=self.__valuetable__[keys[0]]
		else: 
			result={}
			for key,value in self.__valuetable__.items(): 
				if key in keys: result[key]=value
		return result
	def set(self,key,value):
		self.__valuetable__[key]=value
	def __set__(self,key,value):
		self.set(key,value)
	def __getitem__(self,key):
		return self.get(get_table)
	def items(self):
		'''
		if self.__onlock__("Reader"): result=self.__valuetable__.items()
		self.__unlock__("Reader")
		return result
		'''
		return self.__valuetable__.items()
	def update(self,newtable,Operator=None):
		#if Operator=="Reader": return
		if Operator==None: Operator=self.AllocOperator()
		if self.security.acquire(Operator):
			self.__valuetable__.update(newtable)
		self.security.release(Operator)

	def delete(self,*keys,Operator=None):
		#if Operator=="Reader": return
		if Operator==None: Operator=self.AllocOperator()
		if len(keys)==0: keys=self.__valuetable__.keys()
		if self.security.acquire(Operator):
			for key in keys: 
				if key not in ("__STATUS__", "__GUARD__", "__XMEM__"): del self.__valuetable__[key]
		self.security.release(Operator)

	#Monitor Operation
	def get_msg(self,*keys):
		return self.monitor.get(*keys)
	def update_msg(self,newmessage,Operator=None):
		#if Operator=="Reader": return
		if Operator==None: Operator=self.AllocOperator()
		self.monitor.update(newmessage,Operator)
	def delete_msg(self,*keys,Operator=None):
		#if Operator=="Reader": return
		if Operator==None: Operator=self.AllocOperator()
		self.monitor.delete(*keys,Operator)
	def copy_msg(self,Operator=None):
		#if Operator=="Reader": return
		self.delete_msg(Operator)
		self.update_msg(self.__valuetable__,Operator)

	#Operator Name
	def AllocOperator(self,pid=os.getpid()):
		return f"<Operator::{hashlib.md5((str(pid)+str(time.time())).encode('utf-8')).hexdigest()[:self.op_size]}>"

class Guard:
	Status=True
	def __init__(self,ExitTime=10,LoopTime=1e-6,EmpLoopTime=1e-1,VerboseTime=5,script="#Scripts...\n"):
		self.script=script
		self.ExitTime=ExitTime
		self.LoopTime=LoopTime
		self.EmpLoopTime=EmpLoopTime
		self.VerboseTime=VerboseTime
	def __add__(self,newscript):
		self.script+=newscript
	def __Guard__(self,XMEManager,print=print):
		t_a=time.time()
		t_b=time.time()
		t0=time.time()
		while True:
			GuardInfo=XMEManager.get_table("__GUARD__")
			t1=time.time()
			if len(GuardInfo)==0 or "Status" not in GuardInfo or not GuardInfo["Status"] or not self.Status: break
			if XMEManager.len()>0: 
				t_a=t1
				t_b=t1
				XMEManager.exec()
				ex=True
			else: 
				t_b=t1
				ex=False
			if self.VerboseTime!=0 and t1-t0>self.VerboseTime:
				t0=t1
				print(f"XMEManager::Guard:: last heart jump: {round(t_b-t_a,2)} (Dead:> {self.ExitTime}); POC tasks: {XMEManager.len()}; Check time: {self.LoopTime} (empty: {self.EmpLoopTime})")
			if self.ExitTime!=0 and t_b-t_a>self.ExitTime:
				print("XMEManager::Guard:: Time exceeds heartbeat limit, process automatically ends")
				self.Status=False
				break
			time.sleep(self.LoopTime if ex else self.EmpLoopTime)
	def Guard(self, *args, print=print):
		XMEManager=args[-2]
		print=args[-1]
		args=args[:-2]
		Guth=threading.Thread(target=self.__Guard__,args=(XMEManager,print))
		Guth.start()
		__GUARD__=self
		exec(self.script)
		Guth.join()

def __XMEHJ__(status): return status
class HeartJump:
	def __init__(self,Operator,XMEManager,LoopTime=1):
		self.Operator=Operator
		self.Status=True
		self.XMEManager=XMEManager
		self.LoopTime=LoopTime
	def __do__(self):
		while True:
			if not self.Status: break
			self.XMEManager.status(self.Operator)
			time.sleep(self.LoopTime)
	def start(self):
		th=threading.Thread(target=self.__do__)
		th.start()
	def stop(self):
		self.Status=False

def XMEMBuilder(size=32,**initialtable): #automatically build a XMEManager
	sec=Object("Security",Security)
	mon=Object("Monitor",Monitor,sec)
	return Object("XMEManager",XMEManager,sec,mon,initialtable,size)